<?php
/**
*  2020 - Contabilium.com
*
*  @author    Federico Ferreri <federico@ferreri.com.ar>
*  @copyright Contabilium S.A.
*/

class Connector
{
    private $language_id = 1;
    private $api_base = array('AR' => 'https://rest.contabilium.com/', 'UY' => 'https://rest.contabilium.com.uy/', 'CL' => 'https://rest.contabilium.cl/');
    private $country_code = null;
    private $client_id = null;
    private $client_secret = null;
    private $credentials = null;

    private $lastError = null;
    private $lastResult = null;
    private $messages = array();
    private $updatedProducts = array();

    public  $wasQueued = false;

    public function __construct($client_id, $client_secret, $country_code)
    {
        $this->client_id = $client_id;
        $this->client_secret = $client_secret;
        $this->country_code = $country_code;

        if (Configuration::get('CONTABILIUM_TOKEN_CACHE')) {
            $cache = unserialize(Configuration::get('CONTABILIUM_TOKEN_CACHE'));
            if ($cache['expires_ts'] > time()) {
                $this->credentials = $cache['data'];
            }
        }

        $this->language_id = Context::getContext()->language->id;

        return $this;
    }

    public function connect($force = false)
    {
        $this->resetError();

        if ($this->client_id && $this->client_secret) {
            if (!$force && Configuration::get('CONTABILIUM_TOKEN_CACHE')) {
                $cache = unserialize(Configuration::get('CONTABILIUM_TOKEN_CACHE'));
                if ($cache['expires_ts'] > time()) {
                    $this->credentials = $cache['data'];
                    return $this;
                }
            }

            $data = $this->httpPost($this->api_base[$this->country_code] . 'token', array('client_id' => $this->client_id, 'client_secret' => $this->client_secret, 'grant_type' => 'client_credentials'), true);
            if ($data !== false) {
                if (isset($data->error)) {
                    $this->credentials = null;
                    $this->lastError = $data->error;
                } else {
                    $this->credentials = $data;

                    $cache = array(
                        'expires_ts' => time() + $data->expires_in,
                        'data' => $data,
                    );

                    Configuration::updateValue('CONTABILIUM_TOKEN_CACHE', serialize($cache));
                }
            }
        }

        return $this;
    }

    public function isConnected()
    {
        return is_object($this->credentials);
    }

    public function getToken()
    {
        return $this->credentials->access_token;
    }

    public function getLastResult()
    {
        return $this->lastResult;
    }

    public function getLastError()
    {
        return $this->lastError;
    }

    public function hasError()
    {
        return ($this->lastError !== null);
    }

    public function getMessages()
    {
        return $this->messages;
    }

    public function getUpdatedProducts($returnCount = false)
    {
        return ($returnCount) ? count($this->updatedProducts) : $this->updatedProducts;
    }

    public function postOrder($order_id, $is_accepted)
    {
        $this->resetError();
        $this->resetResults();

        if (!Configuration::get('CONTABILIUM_LIVE_MODE')) {
            $this->lastError = 'Integration not enabled.';
            return $this;
        }

        $order = new Order($order_id);
        $orderProducts = $order->getProducts();
        $orderCustomer = $order->getCustomer();
        $orderAddress = new Address($order->id_address_invoice);

        $orderData = array(
            'Cliente' => array(
                'Nombre' => $orderAddress->firstname,
                'Apellido' => $orderAddress->lastname,
                'TipoDocumento' => ($this->country_code == 'UY') ? "CI" : (($this->country_code == 'CL') ? 'RUT' : 'DNI'),
                'Documento' => (trim($orderAddress->dni) !== '') ? preg_replace('/[^0-9]/', '', $orderAddress->dni) : '',
                'Email' => $orderCustomer->email,
                'Telefono' => implode(', ', array($orderAddress->phone, $orderAddress->phone_mobile)),
                'LineaDireccion1' => $orderAddress->address1,
                'LineaDireccion2' => $orderAddress->address2,
                'Ciudad' => $orderAddress->city,
                'Provincia' => State::getNameById($orderAddress->id_state),
                'Pais' => $orderAddress->country,
                'CodigoPostal' => $orderAddress->postcode,
            ),
            'IDVentaIntegracion' => $order_id,
            'IDEstadoIntegracion' => ($is_accepted) ? 'Aceptada' : 'Cancelada',

            'IDIntegracion' => Configuration::get('CONTABILIUM_INTEGRATION_ID'),
            'CondicionVenta' => self::getCondicionVenta($order->reference),
            'FechaEmision' => $order->date_add,
            'Observaciones' => sprintf("Referencia del Pedido: %s\nForma de Pago: %s", $order->reference, $order->payment),
        );

        $orderData['Items'] = array();
        foreach ($orderProducts as $product) {
            $orderData['Items'][] = array(
                'Cantidad' => $product['product_quantity'],
                'Codigo' => $product['product_reference'],
                'Concepto' => $product['product_name'],
                'PrecioUnitario' => $product['unit_price_tax_incl'],
                'Iva' => $product['tax_rate'],
                'Bonificacion' => $product['reduction_amount'],
            );
        }

        // Agrego costo de envío y bonificaciones como un item extra (si aplica)
        if ($order->total_shipping_tax_incl > 0) {
            $carrier = new Carrier($order->id_carrier);

            $orderData['Items'][] = array(
                'Cantidad' => 1,
                'Codigo' => 'CDE',
                'Concepto' => sprintf('Envío (%s)', $carrier->name),
                'PrecioUnitario' => $order->total_shipping_tax_incl,
                'Iva' => $order->carrier_tax_rate,
                'Bonificacion' => 0,
            );
        }

        if ($order->total_discounts_tax_incl > 0) {
            $orderData['Items'][] = array(
                'Cantidad' => 1,
                'Codigo' => '',
                'Concepto' => 'Descuentos / Bonificaciones',
                'PrecioUnitario' => -1 * $order->total_discounts_tax_incl,  // Es un descuento, tiene que restar.
                'Iva' => round(($order->total_discounts_tax_incl  - $order->total_discounts_tax_excl) * 100 / $order->total_discounts_tax_excl, 2),
                'Bonificacion' => 0,
            );
        }

        $this->lastResult = $this->httpPost($this->api_base[$this->country_code] . 'notificador/prestashop', $orderData);

        Db::getInstance()->insert('contabilium_order_tx_log', array(
            'id_order' => $order->id,
            'payload'  => json_encode(array(
                'endpoint' => $this->api_base[$this->country_code] . 'notificador/prestashop',
                'method' => 'POST',
                'data' => $orderData,
                'response' => $this->lastResult
            )),
        ));

        return $this;
    }

    public static function getCondicionVenta($order_reference)
    {
        $condicion_venta = Db::getInstance()->getValue('SELECT payment_method FROM `' . _DB_PREFIX_ . 'order_payment` WHERE order_reference = "' . pSQL($order_reference). '" ORDER BY 1 DESC');
        return ($condicion_venta) ? $condicion_venta : 'Cuenta Corriente';
    }

    public function triggerManualSync()
    {
        $this->resetError();
        $this->updatedProducts = array();

        $productsToSync = $this->getAllProducts();
        if (is_array($productsToSync) && count($productsToSync) > 0) {
            foreach($productsToSync as $item) {
                $this->updateItem($item, true);
            }
        } else if ($productsToSync === -1) {
            $this->wasQueued = true;
            $data = $this->httpPost($this->api_base[$this->country_code] . 'api/conceptos/syncEcommerce?idIntegracion=' . Configuration::get('CONTABILIUM_INTEGRATION_ID'), array());
        }

        return $this;
    }

    public function triggerSync($id)
    {
        $this->resetError();
        $this->updatedProducts = array();

        $item = $this->getSingleProduct($id);
        if ($item && isset($item->Id) && $item->Id > 0) {
            $this->updateItem($item);
        } else {
            $this->lastError = 'No se encontró el producto en Contabilium';
        }
        return $this;
    }

    private function updateItem($item, $isBatch = false)
    {
        $this->messages = array();

        $newPrice = $item->Precio;
        if (Configuration::get('CONTABILIUM_USE_PRICE_WITH_TAX')) {
            $newPrice = $item->PrecioFinal;
        }

        if ($id_product = $this->existsRefInDatabase($item->Codigo)) {
            if ($this->updateProduct($id_product, round($newPrice, 2), $item->Stock, $item->Iva)) {
                $this->updatedProducts[] = $id_product;
            }
        } elseif ($id_combination = $this->existsRefInCombinationsDatabase($item->Codigo)) {
            if ($this->updateProductCombination($id_combination, round($newPrice, 2), $item->Stock)) {
                $this->updatedProducts[] = 'PA:' . $id_combination;
            }
        } else {
            if (!$isBatch) {
                $this->lastError = 'No se encontró el producto en la tienda';
            }
        }
    }

    public function updateProduct($id_product, $newPrice, $newStock, $newTaxRate = null)
    {
        $updated = false;
        $product = new Product($id_product);
        $data_to_log = array('id_product' => $id_product);

        $this->messages[] = sprintf('Producto actualizado: %s', $product->name[$this->language_id]);
        if (Configuration::get('CONTABILIUM_SYNC_PRICE')) {
            $data_to_log['price_old'] = $product->price;
            $product->price = $newPrice;

            // Si el sitio está manejando impuestos entonces
            // empardo el iva seteado en el producto con la tasa que viene de CONTABILIUM
            if (Configuration::get('PS_TAX') && Configuration::get('CONTABILIUM_USE_PRICE_WITH_TAX') && $newTaxRate !== null) {
                // Reglas de Impuestos para Para Argentina
                foreach (TaxRulesGroup::getAssociatedTaxRatesByIdCountry(44) as $tax_rule_id => $rate) {
                    if (intval($rate * 100) === intval($newTaxRate * 100)) {
                        if ($product->id_tax_rules_group !== $tax_rule_id) {
                            $product->id_tax_rules_group = $tax_rule_id;
                        }
                        break;
                    }
                }
            }

            $data_to_log['price_new'] = $product->price;
            $product->update();

            $this->messages[] = sprintf('Precio: %s', $newPrice);
            if (Configuration::get('CONTABILIUM_USE_PRICE_WITH_TAX')) {
                $this->messages[] = 'IVA incluido: SI';
            } else {
                $this->messages[] = 'IVA incluido: NO';
            }

            $updated = true;
        } else {
            $this->messages[] = sprintf('Precio: NO ACTUALIZADO POR CONFIGURACIÓN');
        }

        if (Configuration::get('CONTABILIUM_SYNC_STOCK')) {
            StockAvailable::setQuantity($product->id, null, (int)$newStock);
            $data_to_log['quantity_new'] = $newStock;

            $this->messages[] = sprintf('Stock: %s', (int)$newStock);

            $updated = true;
        } else {
            $this->messages[] = sprintf('Stock: NO ACTUALIZADO POR CONFIGURACIÓN');
        }

        Db::getInstance()->insert('contabilium_prod_update_log', $data_to_log);

        return $updated;
    }

    public function updateProductCombination($id_product_attribute, $newPrice, $newStock)
    {
        $updated = false;
        $row = Db::getInstance()->getRow('SELECT `id_product` FROM `' . _DB_PREFIX_ . 'product_attribute` pa WHERE pa.id_product_attribute = "' . pSQL($id_product_attribute). '"');

        $product = new Product($row['id_product']);

        $data_to_log = array('id_product' => $product->id, 'id_product_attribute' => $id_product_attribute);

        if (Configuration::get('CONTABILIUM_SYNC_PRICE')) {
            $combination = new Combination($id_product_attribute);
            $data_to_log['price_old'] = $combination->price;
            $combination->price = (float)$newPrice;
            $combination->save();
            $data_to_log['price_new'] = $combination->price;

            $updated = true;
        }

        if (Configuration::get('CONTABILIUM_SYNC_STOCK')) {
             StockAvailable::setQuantity($product->id, $id_product_attribute, (int)$newStock);
             $data_to_log['quantity_new'] = $newStock;

             $updated = true;
        }

        return $updated;
    }


    private function getAllProducts()
    {
        $products = array();

        // me traigo la primer página
        $data = $this->httpGet($this->api_base[$this->country_code] . 'api/conceptos/search?idIntegracion=' . Configuration::get('CONTABILIUM_INTEGRATION_ID') . '&filtro=&page=1');
        // guardo los productos
        $products = array_merge($products, $data->Items);

        $totalPages = (int)ceil($data->TotalItems / $data->TotalPage);

        if ($totalPages <= 10) {
            // ahora hago lo mismo de antes pero para el resto de las páginas ($totalPages)
            for($p = 2; $p <= $totalPages; $p++) {
                $data = $this->httpGet($this->api_base[$this->country_code] . 'api/conceptos/search?idIntegracion=' . Configuration::get('CONTABILIUM_INTEGRATION_ID') . '&filtro=&page='.$p);
                $products = array_merge($products, $data->Items);
            };

            return $products;
        } else {
            return -1;
        }
    }

    private function getSingleProduct($id)
    {
        try {
            $id = rawurlencode($id);
            $data = $this->httpGet($this->api_base[$this->country_code] . 'api/conceptos/getInfoForEcommerce?idIntegracion=' . Configuration::get('CONTABILIUM_INTEGRATION_ID') . '&codigo=' . $id);
            if (!isset($data->Id) || $data->Id <= 0) {
                $this->lastError = 'No se encontró el producto.';
            }
        } catch(Exception $e) {
            throw new Exception("Error getting product from Contabilium API [" . $e->getMessage() . ']');
        }

        return $data;
    }

    private function resetError()
    {
        $this->lastError = null;
    }

    private function resetResults()
    {
        $this->lastResult = null;
    }

    private function httpGet($url)
    {
        $this->resetError();

        $token = $this->getAuthToken();

        $ch = curl_init();

        curl_setopt($ch, CURLOPT_HTTPGET, true);
        curl_setopt($ch, CURLOPT_VERBOSE, true);
        curl_setopt($ch, CURLOPT_URL, $url);
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
        curl_setopt($ch, CURLOPT_HEADER, false);
        curl_setopt($ch, CURLOPT_HTTPHEADER, array(
            'Content-Type: application/json',
            'Authorization: Bearer ' . $token
        ));

        $output = curl_exec($ch);
        if ($output) {
            $output = json_decode($output);
        }

        curl_close($ch);
        return $output;
    }

    private function httpPost($url, $params, $isAuthRequest = false, $debug = false)
    {
        $ch = curl_init();

        curl_setopt($ch, CURLOPT_URL, $url);
        curl_setopt($ch, CURLOPT_CUSTOMREQUEST, "POST");
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
        curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, false);
        curl_setopt($ch, CURLOPT_HEADER, $debug);
        curl_setopt($ch, CURLINFO_HEADER_OUT, $debug);

        if (!$isAuthRequest) {
            $token = $this->getAuthToken();

            curl_setopt($ch, CURLOPT_HTTPHEADER, array(
                'Content-Type: application/json',
                'Authorization: Bearer ' . $token
            ));

            $postData = json_encode($params);
        } else {
            $postData = http_build_query($params);
        }

        curl_setopt($ch, CURLOPT_POSTFIELDS, $postData);

        $output = curl_exec($ch);

        if ($output !== false) {
            $output = ($isAuthRequest) ? json_decode($output) : $output;
        } else {
            $this->lastError = curl_error($ch);
        }

        if ($debug) {
            $this->logToFile(sprintf("--- [ BEGIN DEBUG ] ----------\nCURL_INFO:----------\n%s\nOUT: %s\nPOST_DATA:\n%s\n----------\n--- [ END DEBUG ] ----------\n", print_r(curl_getinfo($ch), true), var_export($output, true), print_r($postData, true)));
        }

        curl_close($ch);
        return $output;
    }

    private function getAuthToken()
    {
        $token = false;
        $cached = Configuration::get('CONTABILIUM_TOKEN_CACHE');
        if ($cached !== false) {
            $cached = unserialize(Configuration::get('CONTABILIUM_TOKEN_CACHE'));

            if ($cached['expires_ts'] > time()) {
                $token = $cached['data']->access_token;
            }
        }

        // no hay token en el cache o está expirado, entonces me conecto y consigo uno
        if (!$token) {
            $token = $this->connect(true)->credentials->access_token;
        }

        return $token;
    }

    /**
     * Checks if reference exists
     * @return integer $id_product
     */
    public function existsRefInDatabase($reference)
    {
        $row = Db::getInstance()->getRow('
		SELECT `id_product`
		FROM `' . _DB_PREFIX_ . 'product` p
		WHERE p.reference = "' . pSQL($reference) . '"');

        return $row['id_product'];
    }

    /**
     * Checks if reference exists in product combinations
     * There must be one and only one matching reference, otherwise it can break everyting
     * @return integer $id_combination
     */
    public function existsRefInCombinationsDatabase($reference)
    {
        $rows = Db::getInstance()->executeS('
		SELECT `id_product_attribute`
		FROM `' . _DB_PREFIX_ . 'product_attribute` pa
		WHERE pa.reference = "' . pSQL($reference) . '"');

        return (count($rows) == 1) ? $rows[0]['id_product_attribute'] : false;
    }

    public function logToFile($message)
    {
        if (_PS_MODE_DEV_) {
            file_put_contents(
                __DIR__ . '/../contabilium.log',
                sprintf("[%s] %s\n", date('Y-m-d H:i:s'), $message),
                FILE_APPEND
            );
        }
    }
}